#include "defs.h"
#include <assert.h>

#define _XTAL_FREQ 16000000

void I2C_Initialize(unsigned long freq) {//Begin IIC as master

     TRISC |= 0x18; // Set SDA and SCL pins as input pins
     TRISA.F0 = 0;

     SSPCON  = 0b00101000;    //pg84/234
     SSPCON2 = 0b00000000;    //pg85/234

     SSPADD = _XTAL_FREQ/(4*freq) - 1; // Setting Clock Speed pg99/234
     SSPSTAT = 0b00000000;    //pg83/234

     return;
}//I2C_Initialize

void I2C_Hold() {

// https://simulide.com/p/forum/topic/does-i2c-work-for-any-of-the-pic16f-mcus/
#ifndef SIMULIDE
    delay_us(100);
#else
    while ((SSPCON2 & 0x1F ) || (SSPSTAT & 0x04)); // Wait for any pending transfer
#endif

    return;
}//I2C_Hold

void I2C_Start() {

     I2C_Hold();  //Hold the program is I2C is busy
     SSPCON2.SEN = 1;     //Begin IIC pg85/234
     while(SSPCON2.SEN);

     return;
}//I2C_Start

void I2C_Restart() {

     I2C_Hold();
     SSPCON2.RSEN = 1;
     while(SSPCON2.RSEN);

     return;
}//I2C_Restart

void I2C_Stop() {

     I2C_Hold(); //Hold the program is I2C is busy
     SSPCON2.PEN = 1;    //End IIC pg85/234
     while(SSPCON2.PEN);

     return;
}//I2C_Stop

void I2C_Write(unsigned char datum) {

     I2C_Hold(); // Hold the program is I2C is busy
     SSPBUF = datum; //pg82/234
     while(SSPSTAT.BF);

     I2C_Hold();
     // Check if acknowledge bit has arrived from slave
     if (SSPCON2.ACKSTAT) PORTA.F0 = ~PORTA.F0;

     return;
}//I2C_Write

unsigned char I2C_Read(unsigned char ack) {

         unsigned char incoming;

         I2C_Hold();
         SSPCON2.RCEN = 1;
         while(!SSPSTAT.BF);

         I2C_Hold();
         incoming = SSPBUF;         // get the data saved in SSPBUF

         I2C_Hold();
         SSPCON2.ACKDT = (ack)?0:1; // check if ack bit received
         SSPCON2.ACKEN = 1;         // pg 85/234

         return incoming;
}//I2C_Read

// *****************************

static unsigned char Sensor_get(Sensor* const me) {

       assert(0); // purely-virtual function should never be called

       return 0;  // to avoid compiler warnings
}//Sensor_get

static void Sensor_init(Sensor* const me) {

       assert(0); // purely-virtual function should never be called

       return;    // to avoid compiler warnings
}//Sensor_init

void Sensor_ctor(Sensor* const me) {

     static struct SensorVtbl vtbl;
     vtbl.getData = &Sensor_get;
     vtbl.init = &Sensor_init;
     me->vptr = &vtbl; // "hook" the vptr to the vtbl
     me->I2C_Initialize = &I2C_Initialize;

     return;
}//Sensor_ctor

static unsigned char ds1621_getT(Sensor* const me) {

       ds1621* const _me = (ds1621*)me; // explicit downcast
       unsigned char address = (*_me).WRITE;
       unsigned char tH, tL;

       I2C_Start();
       I2C_Write(address); // Write mode
       I2C_Write(0xAA); // Access temperature register

       I2C_Restart();
       I2C_Write(address | 1); // Read mode
       tH = I2C_Read(1); // Read temperature high, ack
       tL = I2C_Read(0); // Read temperature high, nack
       I2C_Stop();

       return (tH<<8) | tL;
}//ds1621_getT

static void ds1621_init(Sensor* const me) {

       ds1621* const _me = (ds1621*)me; // explicit downcast
       unsigned char address = (*_me).WRITE;

       I2C_Start();
       I2C_Write(address); // Write mode
       I2C_Write(0xAC); // Send configuration register address (Access Config)
       I2C_Write(0x01); // Performs either continuous (0x00) or one-shot (0x01) conversion

       I2C_Restart();
       I2C_Write(address);
       I2C_Write(0xEE); // Send start temperature conversion command
       I2C_Stop();

       return;
}//ds1621_init

void ds1621_ctor(ds1621* const me, unsigned char WRITE) {

     static struct SensorVtbl vtbl; // vtbl of the ds6121 class
     vtbl.getData = &ds1621_getT;
     vtbl.init = &ds1621_init;

     Sensor_ctor(&(*me).super);
     me->super.vptr = &vtbl; // override the vptr

     (*me).WRITE = WRITE;

     return;
}//ds1621_ctor